################################################################################
## calculate_Fpredicted.CDS.pl
##
## Jian-Rong Yang, University of Michigan, Dec. 2014
##
## This code takes a FASTA file, and for each sequence in it, calculates the
## predicted RNA folding strength (F_predicted, or F_p). It also shuffles the
## sequence for 100 times and estimate the expectation of F_p. All results will
## be directly output to STDOUT.
##
## This script assumes the provided FASTA file contains protein coding sequences
## (CDS) only. So please remove any noncoding regions (UTR or intron) before the
## sequences are feed to this script.
##
## This script requires two Perl modules/packages, they are BioPerl (for reading
## FASTA file, and ViennaRNA package (for RNA folding prediction). Note that for
## the ViennaRNA package, you need to enable its Perl module. Please refer to
## the websites of respective tools for installation instructions.
##
##
## For more information about the definition of F_p, please read our paper
##
## Yang, J.-R. and Zhang, J. Human long noncoding RNAs are substantially less
## folded than messenger RNAs. Mol Biol Evol
##
################################################################################

#!/usr/bin/perl -w

use strict;
use Bio::SeqIO;
use RNA;

##
## Read command line arguments.
##
my $fastaFile = shift;

##
## Initialize parameters
##
$RNA::temperature = 37;
my $seqIn = new Bio::SeqIO(-file=>$fastaFile,-format=>"fasta");
my $windowSize = 150;
my $stepSize = 10;
my $nShuffle = 100;

##
## Load sequences
##
my $cnt = 0;
my @allSeq = ();
while(my $seq = $seqIn->next_seq) {
	my $id = $seq->display_id;
	$cnt ++;
	push @allSeq, $seq;
}
print STDERR "Loaded $cnt sequences\n";

sub mean {
	my @val = @_;
	my $sum = 0;
	$sum += $_ for(@val);
	return($sum/(scalar @val));
}

##
## This function does the shuffling of synonymous codons.
## It will handle UTR correctly if you know where the CDS start and end are.
## But when calling this function, the script just assumes that the input FASTA
## file already have UTRs removed.
##
sub getSynCodonShuffled {
	my ($seq,$cdsStart,$cdsEnd) = @_;
	my $residue = ($cdsEnd - $cdsStart + 1) % 3;
	if($residue != 0) {
		$cdsEnd = $cdsEnd - $residue;
	}
	my $utr5 = ($cdsStart > 1)?$seq->subseq(1,$cdsStart):"";
	my $utr3 = ($cdsEnd < $seq->length())?$seq->subseq($cdsEnd+1,$seq->length):"";
	my $cds = $seq->subseq($cdsStart,$cdsEnd);
	my @oriCodons;
	my $aa2codonIdx;
	while($cds =~ /(\w{3})/g) {
		my $thisCodon = $1;
		push @oriCodons,$thisCodon;
		my $thisAa = new Bio::Seq(-display_id=>"foo",-seq=>$thisCodon)->translate->seq;
		if(!exists($aa2codonIdx->{$thisAa})) {
			$aa2codonIdx->{$thisAa} = [scalar @oriCodons - 1];
		} else {
			push @{$aa2codonIdx->{$thisAa}}, scalar @oriCodons - 1;
		}
	}
	my @newCodons = map {""} (1..(scalar @oriCodons));
	foreach my $thisAa (keys %$aa2codonIdx) {
		my @thisIds = @{$aa2codonIdx->{$thisAa}};
		my @newIds = sort {rand() <=> rand()} @thisIds;
		@newCodons[@newIds] = @oriCodons[@thisIds];
	}
	return(new Bio::Seq(-seq=>join("",$utr5,@newCodons,$utr3),display_id=>$seq->display_id));
}

print join("\t","Sequence Name","Observed F_p",map {"Shuffled_".$_." F_p"} (1..$nShuffle))."\n"; ## observation and $nShuffle
foreach my $seq (@allSeq) {
	my $seqId = $seq->display_id();
	my $seqLen = $seq->length();
	my $resPairProb = [map {[]} (1..$seqLen)];
	my $curStart = 1;
	my $curEnd = $curStart + $windowSize - 1;
	my $fName = $seqId;
	$fName =~ s/[^\w\d]/./g;
	while(1) {
		my $frag = $seq->subseq($curStart,$curEnd>$seqLen?$seqLen:$curEnd);
		$frag =~ tr/Tt/UU/;
		my ($struct, $mfe) = RNA::fold($frag);
		my @splitStruct = split("",$struct);
		foreach my $thisId (0..(scalar @splitStruct) - 1) {
			push @{$resPairProb->[$curStart + $thisId - 1]}, (($splitStruct[$thisId]=~/[\.]/)?0:1) ;
		}
		last if($curEnd >= $seqLen);
		$curStart += $stepSize;
		$curEnd = $curStart + $windowSize - 1;
	}
	my $allRes;
	foreach my $thisNt (1..$seqLen) {
		$allRes->{$thisNt} = [&mean(@{$resPairProb->[$thisNt - 1]})];
	}
	## finished with the actual sequence
	
	## now do the shuffled sequence
	foreach my $thisLoop (1..$nShuffle) {
		my $newSeq = &getSynCodonShuffled($seq,1,$seq->length);
		$curStart = 1;
		$curEnd = $curStart + $windowSize - 1;
		$resPairProb = [map {[]} (1..$seqLen)];
		while(1) {
			my $frag = $newSeq->subseq($curStart,$curEnd>$seqLen?$seqLen:$curEnd);
			$frag =~ tr/Tt/UU/;
			my ($struct, $mfe) = RNA::fold($frag);
			my @splitStruct = split("",$struct);
			foreach my $thisId (0..(scalar @splitStruct) - 1) {
				push @{$resPairProb->[$curStart + $thisId - 1]}, (($splitStruct[$thisId]=~/[\.]/)?0:1) ;
			}
			last if($curEnd >= $seqLen);
			$curStart += $stepSize;
			$curEnd = $curStart + $windowSize - 1;
		}
		foreach my $thisNt (1..$seqLen) {
			push @{$allRes->{$thisNt}}, &mean(@{$resPairProb->[$thisNt - 1]});
		}
	}
	my $perSiteRes = {};
	foreach my $thisNt (1..$seqLen) {
		$perSiteRes->{$thisNt} = [@{$allRes->{$thisNt}}];
	}
	my $resProbPerSeq = [];
	foreach my $thisOS (0..$nShuffle) {
		$resProbPerSeq->[$thisOS] = &mean(map {$perSiteRes->{$_}->[$thisOS]} (1..$seqLen));
	}
	print join("\t",$seqId,map {$resProbPerSeq->[$_]} (0..$nShuffle))."\n"; ## observation and $nShuffle
	$cnt --;
	print STDERR "Done with sequence $seqId . $cnt remaining.\n";
}
